#include "ExpressionSimplifier.h"

#include <iostream>
#include "SyFi.h"

using namespace std;
using namespace GiNaC;
using namespace SyFi;


ExpressionSimplifier::ExpressionSimplifier(const string & _tempsymbolbasename, int initialsymbolnumber):
	tempsymbolbasename(_tempsymbolbasename),
	lasttempsymbolnumber(initialsymbolnumber)
{
}


ex ExpressionSimplifier::simplify(ex & e)
{
	e.accept(*this);
	return result;
}

// get the next numbered symbol
const symbol & ExpressionSimplifier::getNextSymbol()
{
	ostringstream name;
	name << tempsymbolbasename << lasttempsymbolnumber;
	lasttempsymbolnumber++;
	const symbol & e = get_symbol(name.str());
	return e;
}

// register an expression in our datastructures and get back a symbol representing it
symbol ExpressionSimplifier::registerExpression(const ex & ein)
{
	ex e = ein.evalf();
	exhashmap<symbol>::iterator i = exToSym.find(e);
	symbol s;
	if( i != exToSym.end() )
	{
		s = i->second;
	}
	else
	{
		s = getNextSymbol();
		exToSym.insert(      pair<ex,symbol>(e, s) );
		symToEx.insert(      pair<symbol,ex>(s, e) );
		symExList.push_back( pair<symbol,ex>(s, e) );
	}
	return s;
}

// helper function encapsulating the "result hack"
const ex & ExpressionSimplifier::doVisit(const ex & e)
{
	e.accept(*this);
	return result;
}


// ... visitors below:

// just treat any unhandled types as a single expression
void ExpressionSimplifier::visit(const basic & e)
{
	result = registerExpression(e);
}

// simplify arguments of functions
void ExpressionSimplifier::visit(const function & e)
{
	ex e2 = e;
	for(int i=0; i<e.nops(); i++)
	{
		ex arg = e.op(i);
		e2 = e2.subs( arg == doVisit(arg) );
	}
	result = registerExpression(e2);
}

// do nothing for the basic types
void ExpressionSimplifier::visit(const numeric & e)
{
	result = e;
}
void ExpressionSimplifier::visit(const constant & e)
{
	// TODO: evalf?
	result = e;
}
void ExpressionSimplifier::visit(const symbol & e)
{
	result = e;
}

// handles integer powers like a series of multiplications,
// and keeps other powers as power() objects
void ExpressionSimplifier::visit(const power & e)
{
	// first visit and simplify both subexpressions
	ex a, b; // a^b

	a = doVisit(e.op(0));
	b = doVisit(e.op(1));

	// if b is integer, treat as n multiplications of a
	ex c = b.evalf();
	if( is_a<numeric>(c) )
	{
		numeric nu = ex_to<numeric>(c);
		int n = (int) to_double(nu);
		if( is_zero(nu-n) )
		{
			// is an integer power, can simplify as a sequence of multiplications
			ex s = a;
			for(int i=0; i<n-1; i++)
			{
				s = registerExpression( s * a );
			}
			result = s;
			return;
		}
	}

	// not integer power, can't simplify more
	result = registerExpression(power(a,b));
}

// splits a sum into a series of single additions
void ExpressionSimplifier::visit(const add & e)
{
	ex s = doVisit(e.op(0));
	for(int i=1; i<e.nops(); i++)
	{
		s = registerExpression( s + doVisit(e.op(i)) );
	}
	result = s;
}

// splits a product into a series of single multiplications 
void ExpressionSimplifier::visit(const mul & e)
{
	ex s = doVisit(e.op(0));
	for(int i=1; i<e.nops(); i++)
	{
		s = registerExpression( s * doVisit(e.op(i)) );
	}
	result = s;
}

