#include <iostream>
using namespace std;

#include <ginac/ginac.h>
using namespace GiNaC;

#include "SyFi.h"
using namespace SyFi;

// for debugging
int exSz(const ex & e)
{
	ostringstream os;
	os << e;
	return os.str().size();
}

void check(const ex & e, const string & name, bool printex=false)
{
	cout << "========================= Expression " << name << endl;
	if(printex) cout << e << endl;
	cout << "string size: " << exSz(e) << endl;
	cout << "has ux to the degrees of ";
	for(int i=0; i<6; i++)
	{
		ex s = get_symbol(istr("u",i));
		int deg  = e.degree(s);
		cout << deg << ", ";
	}
	cout << endl;
	cout << endl;
}

// ... weak form of large deformation elasticity
ex finiteElasticity(const ex & u, const ex & v, const ex & Ginv)
{
	// common constants
	using SyFi::nsd;
	assert(nsd == u.nops());
	ex I = unit_matrix(nsd);

	// TODO: return parameter symbols to code generator somehow, perhaps return pair<ex,lst>?
	ex lambda = get_symbol("lambda");
	ex mu     = get_symbol("mu");

	// unknown derivatives
	ex divU  = div(u, Ginv);
	matrix gradU(nsd,nsd);
	gradU = ex_to<matrix>( (Ginv*grad(u)).evalm() );

	// test function derivatives
	matrix gradV(nsd,nsd);
	gradV = ex_to<matrix>( (Ginv*grad(v)).evalm() );

	cout << "getting started" << endl;

	// deformation gradient
	matrix F(nsd,nsd), FT(nsd,nsd), Finv(nsd,nsd), FTinv(nsd,nsd);
	// TODO: Transpose or not? Different conventions?
	F       = ex_to<matrix>( (I + gradU).evalm() );
	FT      = F.transpose();
	//Finv    = F.inverse();
	FTinv   = FT.inverse();
	ex detF = determinant(F);

	cout << "size F: " << exSz(F) << endl;
	cout << "size FT: " << exSz(F) << endl;
	cout << "size FTinv: " << exSz(F) << endl;
	cout << "size detF: " << exSz(F) << endl;

	//check(F,     "F");
	//check(FT,    "FT");
	//check(FTinv, "FTinv", true);
	//check(detF,  "detF", true);

	// strains
	//matrix C(nsd,nsd), E(nsd,nsd);
	//C = ex_to<matrix>( (FT*F).evalm() );
	//E = ex_to<matrix>( ((C-I)/2).evalm() );

	// stresses
	//   hookes law:
	ex hooke = lambda*divU*I + mu*(gradU + gradU.transpose());
	matrix sigma(nsd,nsd), P(nsd,nsd), S(nsd,nsd);
	sigma = ex_to<matrix>( (hooke).evalm() );
	//   first Piola-Kirchhoff tensor:
	P = ex_to<matrix>( (detF*sigma*FTinv).evalm() ) ;
	//   second Piola-Kirchhoff tensor:
	//S = Finv*P;

	cout << "size hooke: " << exSz(hooke) << endl;
	cout << "size P: " << exSz(P) << endl;

	//check(hooke, "hooke");
	//check(sigma, "sigma", true);
	//check(P, "P = detF*sigma*FTinv");

	// final weak form
	ex e = inner(P, gradV);
	cout << "size e: " << exSz(e) << endl;
	//check(e, "e=P*gradV");
	return e;	
}


// ... weak form of small deformation elasticity
ex smallElasticity(const ex & u, const ex & v, const ex & Ginv)
{
	// common constants
	using SyFi::nsd;
	assert(nsd == u.nops());
	ex I = unit_matrix(nsd);

	// TODO: return parameter symbols to code generator somehow, perhaps return pair<ex,lst>?
	ex lambda = get_symbol("lambda");
	ex mu     = get_symbol("mu");

	// unknown derivatives
	ex divU  = div(u, Ginv);
	matrix gradU(nsd,nsd);
	gradU = ex_to<matrix>( (Ginv*grad(u)).evalm() );

	// test function derivatives
	matrix gradV(nsd,nsd);
	gradV = ex_to<matrix>( (Ginv*grad(v)).evalm() );

	cout << "getting started" << endl;

	// deformation gradient
	matrix F(nsd,nsd), FT(nsd,nsd), Finv(nsd,nsd), FTinv(nsd,nsd);
	// TODO: Transpose or not? Different conventions?
	F       = ex_to<matrix>( (I + gradU).evalm() );
	FT      = F.transpose();
	//Finv    = F.inverse();
	FTinv   = FT.inverse();
	ex detF = determinant(F);

	cout << "size F: " << exSz(F) << endl;
	cout << "size FT: " << exSz(F) << endl;
	cout << "size FTinv: " << exSz(F) << endl;
	cout << "size detF: " << exSz(F) << endl;

	//check(F,     "F");
	//check(FT,    "FT");
	//check(FTinv, "FTinv", true);
	//check(detF,  "detF", true);

	// strains
	//matrix C(nsd,nsd), E(nsd,nsd);
	//C = ex_to<matrix>( (FT*F).evalm() );
	//E = ex_to<matrix>( ((C-I)/2).evalm() );

	// stresses
	//   hookes law:
	ex hooke = lambda*divU*I + mu*(gradU + gradU.transpose());
	matrix sigma(nsd,nsd), P(nsd,nsd), S(nsd,nsd);
	sigma = ex_to<matrix>( (hooke).evalm() );
	//   first Piola-Kirchhoff tensor:
	//P = ex_to<matrix>( (detF*sigma*FTinv).evalm() ) ;
	//   second Piola-Kirchhoff tensor:
	//S = Finv*P;

	cout << "size hooke: " << exSz(hooke) << endl;
	cout << "size P: " << exSz(P) << endl;

	//check(hooke, "hooke");
	//check(sigma, "sigma", true);
	//check(P, "P = detF*sigma*FTinv");

	// final weak form
	ex e = inner(sigma, gradV);
	cout << "size e: " << exSz(e) << endl;
	//check(e, "e=P*gradV");
	return e;	
}
