#include <iostream>
#include <fstream>
#include <string>
using namespace std;

#include <ginac/ginac.h>
using namespace GiNaC;

#include "SyFi.h"
#include "ExpressionSimplifier.h"
using namespace SyFi;


/*
 * Missing features in SyFi base code:
 * - Polygon should be an abstract class. 
 * - Using pow(x,2) is slooow! Should never occur in manually written code!
 */


/*
 * Code generation design plan:
 *
 * List of features:
 * - Choice of 2D/3D and scalar/vector elements only affects the initialization of the code.
 * - Transparent global field to local field dof-handling.
 *
 * Advanced features:
 * - Loading variational forms as "modules" for more flexibility.
 * - Automatic generation of example Python code for user-level solver.
 */

typedef ex (*FormCallback)(const ex & Ni, const ex & Nj, const ex & Ginv);


ex finiteElasticity(const ex & u, const ex & v, const ex & Ginv);
ex smallElasticity(const ex & u, const ex & v, const ex & Ginv);

ex mass(const ex & Ni, const ex & Nj, const ex & Ginv)
{
	ex e = inner(Ni, Nj);
	return e.evalm();
}

ex stiffness(const ex & Ni, const ex & Nj, const ex & Ginv)
{
	ex e = inner( (Ginv*grad(Ni)).evalm(), (Ginv*grad(Nj)).evalm() );
	return e.evalm();
}


inline ex symbolicField(const string & symb, FE & fe)
{
	ex u = 0;
	for(int i=0; i<fe.nbf(); i++)
	{
		u = u + get_symbol(istr(symb,i)) * fe.N(i);
	}
	return u.evalm();
}

string symbolicFieldCode(const string & symb, const FE & fe)
{
	string idofname("iidof");
	ostringstream s;
	s << csrc;
	for(int i=0; i<fe.nbf(); i++)
	{
		s << "    const double " << istr(symb,i) << " = "
		  << symb << "["
		      << idofname << "[" << i << "]"
		  << "];" << endl;
	}
	return s.str();
}


string int2string(int i)
{
	ostringstream os;
	os << i;
	return os.str();
}


// TODO: is this only the linear isotropic mapping?
// We should support more general mappings later,
// this could probably be generated with ginac.
string geometry_string(int nsd)
{
	string s;
	if(nsd == 3)
	{
		s = "\
    double x0 = pp0[0]; double y0 = pp0[1]; double z0 = pp0[2];\n\
    double x1 = pp1[0]; double y1 = pp1[1]; double z1 = pp1[2];\n\
    double x2 = pp2[0]; double y2 = pp2[1]; double z2 = pp2[2];\n\
    double x3 = pp3[0]; double y3 = pp3[1]; double z3 = pp3[2];\n\
 \n\
    double G00 = x1 - x0;\n\
    double G10 = y1 - y0;\n\
    double G20 = z1 - z0;\n\
 \n\
    double G01 = x2 - x0;\n\
    double G11 = y2 - y0;\n\
    double G21 = z2 - z0;\n\
 \n\
    double G02 = x3 - x0;\n\
    double G12 = y3 - y0;\n\
    double G22 = z3 - z0;\n\
 \n\
    double D =  G00*(G11*G22-G21*G12)\n\
              - G01*(G10*G22-G20*G12)\n\
              + G02*(G10*G21-G20*G11);\n";
	}
	else if(nsd == 2)
	{
		s = "\
    double x0 = pp0[0]; double y0 = pp0[1];\n\
    double x1 = pp1[0]; double y1 = pp1[1];\n\
    double x2 = pp2[0]; double y2 = pp2[1];\n\
 \n\
    double G00 = x1 - x0;\n\
    double G01 = x2 - x0;\n\
 \n\
    double G10 = y1 - y0;\n\
    double G11 = y2 - y0;\n\
 \n\
    double D = fabs(G00*G11-G01*G10);\n";
	}
	return s;
}


string create_vector_dof_string(FE & fe, const matrix & G, const matrix & p0, string prefix="i", string dofprefix="")
{
	const int nbf = fe.nbf();
	const int nsd = p0.nops();
	string idof = string("i") + prefix + "dof";
	ostringstream D_str;
	D_str << csrc;
	D_str << "    int    " << idof << "[" << nbf << "];" << endl;
	for(int i=0; i<nbf; i++)
	{
		string sdof = prefix + "dof" + int2string(i);
		string pdof = prefix + "pdof" + int2string(i);

		ex dofi = fe.dof(i);
		ex dofi0 = dofi[0];
		ex dofi1 = dofi[1];
		matrix dofm(1, nsd, ex_to<lst>(dofi0));
		ex gdofm = (G * dofm.transpose() + p0 ).evalm();

		D_str << "    double " << sdof << "[" << (nsd+1) << "];" << endl;
		for(int d=0; d<nsd; d++)
		{
			D_str << "    " << sdof << "[" << d << "] = " << gdofm.op(d) << ";" << endl;
		}
		D_str << "    " << sdof << "[" << nsd << "] = " << dofi1<< ";" << endl; 
		D_str << "    ptv " << pdof << "(" << (nsd+1) << "," << sdof << ");" << endl;
		D_str << "    " << idof << "[" << i << "] = " << dofprefix << "dof.insert_dof(element," << i << "," << pdof << ");" << endl << endl;
	}
	return D_str.str();
}


string create_dof_string(FE & fe, const matrix & G, const matrix & p0, string prefix="i", string dofprefix="")
{
	cerr << "create_dof_string IS NOT TESTED" << endl;

	const int nsd = p0.nops();
	const int nbf = fe.nbf();
	string idof = string("i") + prefix + "dof";
	ostringstream D_str;
	D_str << csrc;
	D_str << "    int    " << idof << "[" << fe.nbf() << "];" << endl;
	for(int i=0; i<fe.nbf(); i++)
	{
		ex dofi  = fe.dof(i);
		matrix dofm(1,nsd);
		dofm = ex_to<matrix>(dofi);
		ex gdofm = (G * dofm.transpose() + p0 ).evalm();
		D_str << "  double dof" << i << "[" << nsd << "];" << endl;
		for(int d=0; i<nsd; i++)
		{
			D_str << "  dof" << i << "[" << d << "]=" << gdofm.op(d) << ";" << endl;
		}
		D_str << endl;
		D_str << "  ptv pdof" << i << "(" << nsd << ",dof" << i << ");" << endl;
		D_str << "  " << idof << "[" << i << "] = " << dofprefix << "dof.insert_dof(element," << i << ",pdof" << i << ");" << endl;
	}
	return D_str.str();
}


// Print code listing for a symbol/expression list.
template<class N, class V>
void genCodeSymbols(ostream & os, string type, const list< pair<N,V> > & slist )
{
	for(typename list< pair<N,V> >::const_iterator i  = slist.begin();
	                                               i != slist.end(); i++)
	{
		os << type << " " << i->first << " = " << i->second << ";" << endl;
	}
}


// for debugging
int exSize(const ex & e)
{
	ostringstream os;
	os << e;
	return os.str().size();
}
		

pair<string,string> generate_code(FE & fe, FormCallback form, string name, bool dotrace = true)
{
	// verbosity control
	ostringstream tmpos;
	ostream & trace = dotrace ? cout : tmpos;

	// get element information 
	const int nbf = fe.nbf();
	Polygon & domain = fe.getPolygon();
	const int nsd = domain.no_space_dim();
	assert(SyFi::nsd == nsd);

	// TODO: get by checking element
	bool vectorElement = true;

	// add space dimension to name
	name = name + int2string(nsd) + "D";
	trace << "Generating code with name " << name << endl;

	// code-generation related variables
	string  filename = string("matrix_") + name;
	string cfilename = filename + ".cc";
	string hfilename = filename + ".hh";
	string symbol = string("__") + filename + "_h__";
	ostringstream code, hcode;
	code  << csrc;
	hcode << csrc;
	string uFieldName = "u";
	
	// write code lead-in
	code << "/*" << endl 
	     << " * Generated by prototype function generate_code." << endl
	     << " */" << endl << endl
	     << "#include \"" << hfilename << "\"" << endl << endl;
	code << "#include <math.h>"   << endl
	     << "#include <utility>"  << endl
	     << "#include <map>"      << endl
	     << "#include \"ptv.h\""  << endl
	     << "#include \"DofT.h\"" << endl
	     << "#include \"Dof_ptv.h\""  << endl
	     << "#include <ScalarFunction.h>" << endl
	     << "#include <VectorFunction.h>" << endl
	     << "" << endl
	     << "using namespace std;"   << endl
	     << "using namespace pycc;" << endl;

	// construct function signature
	ostringstream signature;
	signature << "void " << name << "("
	          << "std::map< std::pair<int,int>, double> & matrix,"
                  << "Dof_ptv & dof, "
                  << "int element";
	for(int i=0; i<nsd+1; i++)
	{
		// TODO: use simplex
		signature  << ", double pp" << i << "[" << nsd << "]";
	}
	signature << ", double *u, int size";
	// TODO: add additional arguments here, for extra symbolic coefficient variables
	signature << ")";
	
	// write function signature
	code  << signature.str() << " {" << endl;

	code << "    const int nbf = " << nbf << ";" << endl << endl;
	
	// symbolic geometric mapping
	matrix G(nsd,nsd);
	G = ex_to<matrix>( symbolic_matrix(nsd,nsd,"G") );
	ex x0 = get_symbol("x0");
	ex y0 = get_symbol("y0") ;
	ex z0 = get_symbol("z0") ;
	matrix p0(nsd,1);
	if(nsd == 2)
	{
		p0 = x0, y0;
	}
	else if(nsd == 3)
	{
		p0 = x0, y0, z0;
	}
	ex Ginv = G.inverse().transpose();
	ex D = get_symbol("D");
	// symbolic representative of the inverse mapping
	matrix Jinv(nsd,nsd);
	Jinv = ex_to<matrix>( symbolic_matrix(nsd,nsd,"Jinv") );

	// write geometry string
  	string gs = geometry_string(nsd);
  	code << gs << endl;
	// write code for "Jinv = Ginv"
	for(int i=0; i<Jinv.nops(); i++)
	{
		code << "    const double " << Jinv.op(i) << " = " << Ginv.op(i) << ";" << endl;
	}
	code << endl;

	// create dofstring from (fe,G,p0)
	string dof_string;
	if(vectorElement)
	{
		dof_string = create_vector_dof_string(fe, G, p0);
	}
	else
	{
		dof_string = create_dof_string(fe, G, p0);
	}
	code << dof_string << endl;


	// TODO: handle extra symbolic coefficient variables like "c = symbol("c")"

	trace << "Calculating element integrals (could take some time)." << endl;
	// calculate element integral, simplify, and write element matrix code
	ExpressionSimplifier es;
	ostringstream A_str;
	A_str << csrc;
	A_str << "    double A[nbf][nbf];" << endl;

	// u = sum uk Nk
	ex u = symbolicField(uFieldName, fe);
	string sfc = symbolicFieldCode(uFieldName, fe);
	code << sfc << endl;

	// for each equation / test function
	for(int j=0; j<nbf; j++)
	{
		trace << "Handling equation " << j << endl;

		// integrate equation j
		ex integrand = D*form(u, fe.N(j), Jinv);	// TODO: handle extra symbolic coefficient variables
		trace << "integrand size: " << exSize(integrand) << ", ";
		integrand    = normal(integrand);
		trace << exSize(integrand) << endl;

		ex Aj = domain.integrate(integrand);
		trace << "Aj size: " << exSize(Aj) << ", ";
		Aj = normal(Aj);
		trace << exSize(Aj) << endl;

		// calculate (Jacobi) matrix entries
		for(int i=0; i<nbf; i++)
		{
			// differentiate wrt ui
			ex Aij  = Aj.diff( get_symbol(istr("u",i)) );
			trace << "Aij size: " << exSize(Aij) << ", ";
			// normalize expression
			Aij = normal(Aij).evalf();
			trace << exSize(Aij) << endl;
			// extract reusable expression parts
			ex sAij = es.simplify(Aij);
			// write code
			A_str << "    A[" << i << "][" << j << "] = " << sAij << ";" << endl;
		}
	}
	trace << "Writing element matrix code (could be large)." << endl;
	// write code for all temporary variables in es
	genCodeSymbols(code, "    const double", es.getSymExList());
	// write code for simplified matrix entries
	code << A_str.str() << endl;


	trace << "Writing lead-out of element matrix code." << endl;
	// write code to insert local element matrix into global matrix
	code << "    pair<int,int> index;"               << endl
	     << "    for (int i=0; i< nbf; i++) {"       << endl
	     << "        index.first = iidof[i];"        << endl
	     << "        for (int j=0; j< nbf; j++) {"   << endl
	     << "            index.second = iidof[j];"   << endl
	     << "            matrix[index] += A[i][j];"  << endl
	     << "        }" << endl
	     << "    }" << endl;
	// write code leadout
	code <<  "}" << endl;

	// write header code
	hcode << "/*" << endl 
	      << " * Generated by prototype function generate_code." << endl
	      << " */" << endl
	      << endl
	      << "#ifndef " << symbol << endl
	      << "#define " << symbol << endl
	      << endl
	      << "#include <utility>"  << endl
	      << "#include <map>"      << endl
	      << "#include \"ptv.h\""  << endl
	      << "#include \"DofT.h\"" << endl
	      << "#include \"Dof_ptv.h\""  << endl
	      << endl
	      << signature.str() << ";" << endl
	      << endl
	      << "#endif // #ifndef " << symbol << endl;

	// write code to files
	string s = code.str();
	string hs = hcode.str();
	if(name != "")
	{
		trace << "Writing code to files." << endl;

		ofstream f( cfilename.c_str() );
		f << s;
		f.close();

		ofstream h( hfilename.c_str() );
		h << hs;
		h.close();
	}

	// return header code and source code
	return make_pair(hs, s);
}

int main(int argc, char ** argv)
{
	bool verbose = true;

	// setup a reference polygon
	int nsd = 2;
	SyFi::nsd = nsd;
	ReferenceTriangle domain;
	//ReferenceTetrahedron domain;

	// setup a finite element
	int order = 1;
	VectorLagrangeFE fe;
	fe.set_size(nsd);
	fe.set(domain);
	fe.set(order);
	fe.compute_basis_functions();

	cout << fe.nbf() << endl;

	// generate C++ code for computing the element matrix
	//pair<string,string> ps = generate_code(fe, mass, "mass", verbose);
	//pair<string,string> ps = generate_code(fe, stiffness, "stiffness", verbose);
	//pair<string,string> ps = generate_code(fe, finiteElasticity, "finiteElasticity", verbose);
	pair<string,string> ps = generate_code(fe, smallElasticity, "smallElasticity", verbose);

	/*
	if(verbose)
	{
		cout << "...... Header code:" << endl;
		cout << ps.first << endl;
		cout << "...... Source code:" << endl;
		cout << ps.second << endl;
	}
	*/
	return 0;
}

